/***********************************************************************************************************************
 * Copyright [2018] RELOC s.r.l. - All rights strictly reserved
 * This Software is provided for evaluation purposes; any other use — including reproduction, modification, use for
 * a commercial product, distribution, or republication — without the prior written permission of the Copyright holder
 * is strictly prohibited. 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS
 * OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 **********************************************************************************************************************/

/**********************************************************************************************************************
 * File Name        : sf_ili9225.c
 * Description      : ILI9225 TFT LCD Single Chip Driver.
 *********************************************************************************************************************/


/**********************************************************************************************************************
 * Includes
 *********************************************************************************************************************/
#include "sf_ili9225.h"


/**********************************************************************************************************************
 * Macro definitions
 *********************************************************************************************************************/
#define MAX( a, b )                                         ((a) > (b) ? (a) : (b))
#define MIN( a, b )                                         ((a) > (b) ? (b) : (a))
#define CONV_MS_TO_TICK(a)                                  ((a>0) ? MAX( ((a)*TX_TIMER_TICKS_PER_SECOND/1000), 1 ) : 0)
#define SF_ILI9225_ASSERT( a )                              if( (a) != SSP_SUCCESS ) return a

#define SF_ILI9225_INSTANCE_OPEN                            0xFB7AA3F6
#define SF_ILI9225_CMD_REG_LEN                              1
#define SF_ILI9225_DATA_LEN                                 2
#define SPI_MAX_TRANSFER_LENGTH                             0xFFFF

#define SF_ILI9225_CHIP_ID_VALUE                            0x9225
#define SF_ILI9225_REG_DRIVER_OUTPUT_CONTROL_ROTATION_0     0x0100
#define SF_ILI9225_REG_DRIVER_OUTPUT_CONTROL_ROTATION_180   0x0200


#define SF_ILI9225_DELAY                                    0xFF
#define SF_ILI9225_REG_CHIP_ID_CODE                         0x00
#define SF_ILI9225_REG_DRIVER_OUTPUT_CONTROL                0x01
#define SF_ILI9225_REG_LCD_AC_DRIVING_CONTROL               0x02
#define SF_ILI9225_REG_ENTRY_MODE                           0x03
#define SF_ILI9225_REG_DISPLAY_CONTROL_1                    0x07
#define SF_ILI9225_REG_BLANK_PERIOD_CONTROL_1               0x08
#define SF_ILI9225_REG_FRAME_CYCLE_CONTROL                  0x0B
#define SF_ILI9225_REG_INTERFACE_CONTROL                    0x0C
#define SF_ILI9225_REG_OSCILLATION_CONTROL                  0x0F
#define SF_ILI9225_REG_POWER_CONTROL_1                      0x10
#define SF_ILI9225_REG_POWER_CONTROL_2                      0x11
#define SF_ILI9225_REG_POWER_CONTROL_3                      0x12
#define SF_ILI9225_REG_POWER_CONTROL_4                      0x13
#define SF_ILI9225_REG_POWER_CONTROL_5                      0x14
#define SF_ILI9225_REG_RAM_ADDRESS_SET_1                    0x20
#define SF_ILI9225_REG_RAM_ADDRESS_SET_2                    0x21
#define SF_ILI9225_REG_WRITE_DATA_TO_GRAM                   0x22
#define SF_ILI9225_REG_SOFTWARE_RESET                       0x28
#define SF_ILI9225_REG_GATE_SCAN_CONTROL                    0x30
#define SF_ILI9225_REG_VERTICAL_SCROLL_CONTROL_1            0x31
#define SF_ILI9225_REG_VERTICAL_SCROLL_CONTROL_2            0x32
#define SF_ILI9225_REG_VERTICAL_SCROLL_CONTROL_3            0x33
#define SF_ILI9225_REG_PARTIAL_DRIVING_POSITION_1           0x34
#define SF_ILI9225_REG_PARTIAL_DRIVING_POSITION_2           0x35
#define SF_ILI9225_REG_HORIZONTAL_WINDOW_ADDRESS_1          0x36
#define SF_ILI9225_REG_HORIZONTAL_WINDOW_ADDRESS_2          0x37
#define SF_ILI9225_REG_VERTICAL_WINDOW_ADDRESS_1            0x38
#define SF_ILI9225_REG_VERTICAL_WINDOW_ADDRESS_2            0x39
#define SF_ILI9225_REG_GAMMA_CONTROL_1                      0x50
#define SF_ILI9225_REG_GAMMA_CONTROL_2                      0x51
#define SF_ILI9225_REG_GAMMA_CONTROL_3                      0x52
#define SF_ILI9225_REG_GAMMA_CONTROL_4                      0x53
#define SF_ILI9225_REG_GAMMA_CONTROL_5                      0x54
#define SF_ILI9225_REG_GAMMA_CONTROL_6                      0x55
#define SF_ILI9225_REG_GAMMA_CONTROL_7                      0x56
#define SF_ILI9225_REG_GAMMA_CONTROL_8                      0x57
#define SF_ILI9225_REG_GAMMA_CONTROL_9                      0x58
#define SF_ILI9225_REG_GAMMA_CONTROL_10                     0x59
#define SF_ILI9225_REG_NV_MEMORY_DATA_PROGRAMMING           0x60
#define SF_ILI9225_REG_NV_MEMORY_CONTROL                    0x61
#define SF_ILI9225_REG_NV_MEMORY_STATUS                     0x62
#define SF_ILI9225_REG_NV_MEMORY_PROTECTION_KEY             0x63
#define SF_ILI9225_REG_ID_CODE                              0x65
#define SF_ILI9225_REG_SPI_READ_WRITE_CONTROL               0x66

typedef struct st_sf_ili9225_cmd
{
    uint8_t command;
    uint16_t data;
} sf_ili9225_cmd;

/***********************************************************************************************************************
 * Public Instance API Functions
 **********************************************************************************************************************/
ssp_err_t SF_ILI9225_Open ( sf_ili9225_ctrl_t * const p_ctrl,
                            sf_ili9225_cfg_t const * const p_cfg );
ssp_err_t SF_ILI9225_Close ( sf_ili9225_ctrl_t * const p_ctrl );
ssp_err_t SF_ILI9225_DrawFull ( sf_ili9225_ctrl_t * const p_ctrl,
                                uint8_t* p_src );
ssp_err_t SF_ILI9225_DrawPartial ( sf_ili9225_ctrl_t * const p_ctrl,
                                   uint8_t* p_src,
                                   uint16_t x_offset,
                                   uint16_t y_offset,
                                   uint16_t row_len,
                                   uint16_t rows);

/**********************************************************************************************************************
 * Private function prototypes
 *********************************************************************************************************************/
static ssp_err_t SF_ILI9225_WriteReg( sf_ili9225_ctrl_t * const p_ctrl,
                                      uint8_t reg_addr,
                                      uint16_t reg_value );
static ssp_err_t SF_ILI9225_WriteGRAM( sf_ili9225_ctrl_t * const p_ctrl,
                                       uint8_t* data,
                                       uint32_t data_len );

/**********************************************************************************************************************
 * Public global variables
 *********************************************************************************************************************/
sf_ili9225_api_t sf_ili9225_api =
{
    .open = SF_ILI9225_Open,
    .close = SF_ILI9225_Close,
    .drawFull = SF_ILI9225_DrawFull,
    .drawPartial = SF_ILI9225_DrawPartial,
};

/**********************************************************************************************************************
 * Private global variables
 *********************************************************************************************************************/
static const sf_ili9225_cmd init_cmd_list[] =
{
    /* Initial Sequence */
    { SF_ILI9225_REG_DRIVER_OUTPUT_CONTROL,         0x001C },       // NL (528*220 dots) (Screen orientation depending on configuration)
    { SF_ILI9225_REG_LCD_AC_DRIVING_CONTROL,        0x0100 },       // 1 Line Inversion
    { SF_ILI9225_REG_ENTRY_MODE,                    0x1030 },       // GRAM update direction
    { SF_ILI9225_REG_BLANK_PERIOD_CONTROL_1,        0x0808 },       // Back Porch = 8 lines, Front Porch = 8 lines
    { SF_ILI9225_REG_INTERFACE_CONTROL,             0x0001 },       // RGB 16bit
    { SF_ILI9225_REG_OSCILLATION_CONTROL,           0x0801 },       // Refresh Rate 71 Hz
    { SF_ILI9225_REG_RAM_ADDRESS_SET_1,             0x0000 },       // Initial RAM address
    { SF_ILI9225_REG_RAM_ADDRESS_SET_2,             0x0000 },       // Initial RAM address
    /* Power On Sequence */
    { SF_ILI9225_DELAY,                             50     },
    { SF_ILI9225_REG_POWER_CONTROL_1,               0x0A00 },       // Gamma amplification to default
    { SF_ILI9225_REG_POWER_CONTROL_2,               0x103B },       // VCI1 at 2.76 V
    { SF_ILI9225_DELAY,                             50     },
    { SF_ILI9225_REG_POWER_CONTROL_3,               0x3121 },       // VGH = 5x VCI1, VGL = -4x VC1, StepUp1= 1/2 Fosc, StepUp2= 1/16 Fosc, StepUp3= 1/8 Fosc
    { SF_ILI9225_REG_POWER_CONTROL_4,               0x0066 },       // GVDD = Vreg1out = 4.5 V
    { SF_ILI9225_REG_POWER_CONTROL_5,               0x3660 },       // VCOML = GVDD * 0.6985, VCOMH = GVDD * 1.014
    /* Set GRAM area */
    { SF_ILI9225_REG_GATE_SCAN_CONTROL,             0x0000 },       // Start from G1
    { SF_ILI9225_REG_VERTICAL_SCROLL_CONTROL_1,     0x00DB },       // End from 219
    { SF_ILI9225_REG_VERTICAL_SCROLL_CONTROL_2,     0x0000 },       // Start from 0
    { SF_ILI9225_REG_VERTICAL_SCROLL_CONTROL_3,     0x0000 },       // Step
    { SF_ILI9225_REG_PARTIAL_DRIVING_POSITION_1,    0x00DB },       // End from 219
    { SF_ILI9225_REG_PARTIAL_DRIVING_POSITION_2,    0x0000 },       // Start from 0
    { SF_ILI9225_REG_HORIZONTAL_WINDOW_ADDRESS_1,   0x00AF },       // End from 175
    { SF_ILI9225_REG_HORIZONTAL_WINDOW_ADDRESS_2,   0x0000 },       // Start from 0
    { SF_ILI9225_REG_VERTICAL_WINDOW_ADDRESS_1,     0x00DB },       // End from 219
    { SF_ILI9225_REG_VERTICAL_WINDOW_ADDRESS_2,     0x0000 },       // Start from 0
    /* Adjust Gamma Curve */
    { SF_ILI9225_REG_GAMMA_CONTROL_1,               0x0400 },       // Gamma
    { SF_ILI9225_REG_GAMMA_CONTROL_2,               0x080B },       // Gamma
    { SF_ILI9225_REG_GAMMA_CONTROL_3,               0x0E0C },       // Gamma
    { SF_ILI9225_REG_GAMMA_CONTROL_4,               0x0103 },       // Gamma
    { SF_ILI9225_REG_GAMMA_CONTROL_5,               0x0C0E },       // Gamma
    { SF_ILI9225_REG_GAMMA_CONTROL_6,               0x0B08 },       // Gamma
    { SF_ILI9225_REG_GAMMA_CONTROL_7,               0x0004 },       // Gamma
    { SF_ILI9225_REG_GAMMA_CONTROL_8,               0x0301 },       // Gamma
    { SF_ILI9225_REG_GAMMA_CONTROL_9,               0x0E00 },       // Gamma
    { SF_ILI9225_REG_GAMMA_CONTROL_10,              0x000E },       // Gamma
    /* Turn on display */
    { SF_ILI9225_DELAY,                             50     },
    { SF_ILI9225_REG_DISPLAY_CONTROL_1,             0x1017 },       // Activate display
};

/**********************************************************************************************************************
 * Functions
 *********************************************************************************************************************/
ssp_err_t SF_ILI9225_Open ( sf_ili9225_ctrl_t * const p_ctrl,
                            sf_ili9225_cfg_t const * const p_cfg )
{
    ssp_err_t status;
    uint16_t rotation_cfg;
    
    p_ctrl->cmd_pin = p_cfg->cmd_pin;
    p_ctrl->reset_pin = p_cfg->reset_pin;
    p_ctrl->spi = p_cfg->spi;
    p_ctrl->x_size = p_cfg->x_size;
    p_ctrl->y_size = p_cfg->y_size;
    p_ctrl->byte_per_pixel = p_cfg->byte_per_pixel;
    
    switch( p_cfg->rotation_angle )
    {
        case 0  :   rotation_cfg = SF_ILI9225_REG_DRIVER_OUTPUT_CONTROL_ROTATION_0;     break;
        case 180:   rotation_cfg = SF_ILI9225_REG_DRIVER_OUTPUT_CONTROL_ROTATION_180;   break;
        default :   SF_ILI9225_ASSERT( SSP_ERR_INVALID_ARGUMENT );                      break;
    }
    
    status = p_ctrl->spi->p_api->open(p_ctrl->spi->p_ctrl, p_ctrl->spi->p_cfg);
    
    g_ioport.p_api->pinWrite(p_ctrl->reset_pin, IOPORT_LEVEL_HIGH);
    g_ioport.p_api->pinWrite(p_ctrl->cmd_pin, IOPORT_LEVEL_HIGH);
    
    if (SSP_SUCCESS == status)
    {
        uint32_t i;
        
        /* Reset cycle */
        tx_thread_sleep(CONV_MS_TO_TICK(1));
        g_ioport.p_api->pinWrite(p_ctrl->reset_pin, IOPORT_LEVEL_LOW);
        tx_thread_sleep(CONV_MS_TO_TICK(10));
        g_ioport.p_api->pinWrite(p_ctrl->reset_pin, IOPORT_LEVEL_HIGH);
        tx_thread_sleep(CONV_MS_TO_TICK(50));
        
        status = p_ctrl->spi->p_api->lock( p_ctrl->spi->p_ctrl );
        
        if( status == SSP_SUCCESS )
        {
            for( i=0 ; i<(sizeof(init_cmd_list)/sizeof(sf_ili9225_cmd)) ; i++ )
            {
                if( init_cmd_list[i].command == SF_ILI9225_DELAY )
                {
                    /* Sleep command */
                    tx_thread_sleep(CONV_MS_TO_TICK(init_cmd_list[i].data));
                }
                else if ( init_cmd_list[i].command == SF_ILI9225_REG_DRIVER_OUTPUT_CONTROL )
                {
                    /* Add orientation configuration */
                    status = SF_ILI9225_WriteReg( p_ctrl, init_cmd_list[i].command, (init_cmd_list[i].data | rotation_cfg) );
                    if(SSP_SUCCESS != status)
                    {
                        break;
                    }
                }
                else
                {
                    /* Normal command */
                    status = SF_ILI9225_WriteReg( p_ctrl, init_cmd_list[i].command, init_cmd_list[i].data );
                    if(SSP_SUCCESS != status)
                    {
                        break;
                    }
                }
            }
            
            p_ctrl->spi->p_api->unlock( p_ctrl->spi->p_ctrl );
        }
    }
    
    if (SSP_SUCCESS == status)
    {
        p_ctrl->open = SF_ILI9225_INSTANCE_OPEN;
    }
    
    return status;
}


ssp_err_t SF_ILI9225_Close (sf_ili9225_ctrl_t * const p_ctrl)
{
    ssp_err_t status = SSP_SUCCESS;
    
    if( p_ctrl->open == SF_ILI9225_INSTANCE_OPEN )
    {
        status = p_ctrl->spi->p_api->close(p_ctrl->spi->p_ctrl);
        p_ctrl->open = 0;
    }
    
    return status;
}


ssp_err_t SF_ILI9225_DrawFull (sf_ili9225_ctrl_t * const p_ctrl,
                               uint8_t* p_src)
{
    return SF_ILI9225_DrawPartial ( p_ctrl,
                                    p_src,
                                    0,                  /* X Offset */
                                    0,                  /* Y Offset */
                                    p_ctrl->x_size,     /* Row length */
                                    p_ctrl->y_size );   /* Rows number */
}

ssp_err_t SF_ILI9225_DrawPartial (sf_ili9225_ctrl_t * const p_ctrl,
                                  uint8_t * p_src,
                                  uint16_t x_offset,
                                  uint16_t y_offset,
                                  uint16_t row_len,
                                  uint16_t rows)
{
    uint32_t total_bytes = (uint32_t) row_len*rows*p_ctrl->byte_per_pixel;
    ssp_err_t status;
    status = p_ctrl->spi->p_api->lock( p_ctrl->spi->p_ctrl );
    
    if( status == SSP_SUCCESS )
    {
        /* Configure window */
        SF_ILI9225_ASSERT( SF_ILI9225_WriteReg( p_ctrl, SF_ILI9225_REG_HORIZONTAL_WINDOW_ADDRESS_1, (uint16_t)(x_offset + row_len - 1) ) );
        SF_ILI9225_ASSERT( SF_ILI9225_WriteReg( p_ctrl, SF_ILI9225_REG_HORIZONTAL_WINDOW_ADDRESS_2, x_offset ) );
        SF_ILI9225_ASSERT( SF_ILI9225_WriteReg( p_ctrl, SF_ILI9225_REG_VERTICAL_WINDOW_ADDRESS_1, (uint16_t)(y_offset + rows - 1) ) );
        SF_ILI9225_ASSERT( SF_ILI9225_WriteReg( p_ctrl, SF_ILI9225_REG_VERTICAL_WINDOW_ADDRESS_2, y_offset ) );
        
        /* Set initial address */
        SF_ILI9225_ASSERT( SF_ILI9225_WriteReg( p_ctrl, SF_ILI9225_REG_RAM_ADDRESS_SET_1, x_offset ) );
        SF_ILI9225_ASSERT( SF_ILI9225_WriteReg( p_ctrl, SF_ILI9225_REG_RAM_ADDRESS_SET_2, y_offset ) );
        
        /* Send data */
        SF_ILI9225_ASSERT( SF_ILI9225_WriteGRAM( p_ctrl, p_src, (uint32_t)total_bytes ) );
        
        p_ctrl->spi->p_api->unlock( p_ctrl->spi->p_ctrl );
    }
    return status;
}

static ssp_err_t SF_ILI9225_WriteReg( sf_ili9225_ctrl_t * const p_ctrl,
                                      uint8_t reg_addr,
                                      uint16_t reg_value )
{
    ssp_err_t status;
    
    /* Send command first */
    g_ioport.p_api->pinWrite(p_ctrl->cmd_pin, IOPORT_LEVEL_LOW);
    status = p_ctrl->spi->p_api->write(p_ctrl->spi->p_ctrl, &reg_addr, SF_ILI9225_CMD_REG_LEN, SPI_BIT_WIDTH_8_BITS, TX_WAIT_FOREVER);
    g_ioport.p_api->pinWrite(p_ctrl->cmd_pin, IOPORT_LEVEL_HIGH);
    if (SSP_SUCCESS == status)
    {
        /* Send data using a support buffer for having the correct byte order */
        uint8_t tmp_buf[SF_ILI9225_DATA_LEN];
        tmp_buf[0] = (uint8_t) (reg_value >> 8);
        tmp_buf[1] = (uint8_t) (reg_value & 0xFF);
        status = p_ctrl->spi->p_api->write(p_ctrl->spi->p_ctrl, &tmp_buf[0], SF_ILI9225_DATA_LEN, SPI_BIT_WIDTH_8_BITS, TX_WAIT_FOREVER);
    }
    return status;
}

static ssp_err_t SF_ILI9225_WriteGRAM( sf_ili9225_ctrl_t * const p_ctrl,
                                       uint8_t* data,
                                       uint32_t data_len )
{
    ssp_err_t status;
    uint8_t tmp_buf[SF_ILI9225_CMD_REG_LEN];
    tmp_buf[0] = SF_ILI9225_REG_WRITE_DATA_TO_GRAM;
    /* Send command first */
    g_ioport.p_api->pinWrite(p_ctrl->cmd_pin, IOPORT_LEVEL_LOW);
    status = p_ctrl->spi->p_api->write(p_ctrl->spi->p_ctrl, &tmp_buf[0], SF_ILI9225_CMD_REG_LEN, SPI_BIT_WIDTH_8_BITS, TX_WAIT_FOREVER);
    g_ioport.p_api->pinWrite(p_ctrl->cmd_pin, IOPORT_LEVEL_HIGH);
    if (SSP_SUCCESS == status)
    {
        uint32_t data_to_send;
        uint32_t data_sent = 0;
        
        /* Send data (splitting it in segments if too long the SPI */
        p_ctrl->spi->p_api->lock( p_ctrl->spi->p_ctrl );
        while( data_sent < data_len )
        {
            data_to_send = MIN( (data_len-data_sent), SPI_MAX_TRANSFER_LENGTH );
            status = p_ctrl->spi->p_api->write(p_ctrl->spi->p_ctrl, &data[data_sent], data_to_send, SPI_BIT_WIDTH_8_BITS, TX_WAIT_FOREVER);
            data_sent += data_to_send;
        }
        p_ctrl->spi->p_api->unlock( p_ctrl->spi->p_ctrl );
    }
    return status;
}
